# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command line entry point for calling trend_analyzer."""

import argparse

from gaarf.cli import utils as gaarf_utils
from gaarf.io import writer

from trend_analyzer import analyzer, llms, vectorstore


def main():  # noqa: D103
  parser = argparse.ArgumentParser()
  parser.add_argument('question')
  parser.add_argument('--llm', dest='llm')
  parser.add_argument('--output-type', dest='output', default='console')
  parser.add_argument(
    '--output-destination', dest='destination', default='results'
  )
  args, kwargs = parser.parse_known_args()
  params = gaarf_utils.ParamsParser(['llm', args.output]).parse(kwargs)

  trend_analyzer_llm = llms.create_trend_analyzer_llm(
    args.llm, params.get('llm')
  )
  trend_analyze_handler = analyzer.TrendsAnalyzer(
    vect_store=vectorstore.load_vectorstore(), llm=trend_analyzer_llm
  )
  result = trend_analyze_handler.analyze(args.question)
  writer.create_writer(args.output, **params.get(args.output)).write(
    result, args.destination
  )


if __name__ == '__main__':
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining various LLMs for performing trends analysis."""

from __future__ import annotations

import langchain_google_vertexai
from langchain_core import language_models


def create_trend_analyzer_llm(
  llm_type: str, llm_parameters: dict[str, str] | None = None
) -> language_models.BaseLanguageModel:
  """Creates LLM based on type and parameters.

  Args:
    llm_type: Type of LLM to instantiate.
    llm_parameters: Various parameters to instantiate LLM.

  Returns:
    Initialized LLM.

  Raises:
    TrendsAnalyzerLLMError: When incorrect LLM type is specified.
  """
  mapping = {
    'gemini': langchain_google_vertexai.ChatVertexAI,
    'fake': language_models.FakeListLLM,
  }
  if llm := mapping.get(llm_type):
    if not llm_parameters:
      llm_parameters = {}
    return llm(**llm_parameters)
  raise TrendsAnalyzerLLMError(f'Unsupported LLM type: {llm_type}')


class TrendsAnalyzerLLMError(Exception):
  """Error when incorrect LLM type is specified."""



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for performing trends analysis.

Exposes a single class TrendAnalyzer that allows to get extract trends from
vector store based on provided LLM.
"""

from __future__ import annotations

import pathlib

import gaarf
import smart_open
from langchain_core import (
  language_models,
  output_parsers,
  prompts,
  pydantic_v1,
  runnables,
  vectorstores,
)


class Trend(pydantic_v1.BaseModel):
  """Defines structured output for LLM.

  Attributes:
    name: Name of the trend.
    growth: Relative growth associated with trend.
  """

  name: str = pydantic_v1.Field(description='trend name')
  growth: float = pydantic_v1.Field(description='trend growth')


def _format_docs(docs):
  return '\n\n'.join(doc.page_content for doc in docs)


class TrendsAnalyzer:
  """Handles prompts related to analyzing trends.

  Attributes:
    vect_store: Vector store containing data on trends.
    llm: LLM responsible to handle prompts.
  """

  def __init__(
    self,
    vect_store: vectorstores.VectorStore,
    llm: language_models.BaseLanguageModel,
  ) -> None:
    """Initializes TrendsAnalyzer based on vectorstore and LLM.

    Args:
      vect_store: Vector store containing data on trends.
      llm: LLM responsible to handle prompts.
    """
    self.vect_store = vect_store
    self.llm = llm

  @property
  def prompt(self) -> prompts.PromptTemplate:
    """Builds correct prompt to send to LLM.

    Prompt contains format instructions to get output result.
    """
    with smart_open.open(
      pathlib.Path(__file__).resolve().parent / 'template.txt',
      'r',
      encoding='utf-8',
    ) as f:
      template = f.readlines()
    return prompts.PromptTemplate(
      template=' '.join(template),
      input_variables=['question'],
      partial_variables={
        'format_instructions': self.output_parser.get_format_instructions(),
      },
    )

  @property
  def output_parser(self) -> output_parsers.BaseOutputParser:
    """Defines how LLM response should be formatted."""
    return output_parsers.JsonOutputParser(pydantic_object=Trend)

  def analyze(self, text: str) -> gaarf.report.GaarfReport:
    """Performs trend analysis based on a provided question.

    Args:
      text: Question to LLM.

    Returns:
      Report with trends data.
    """
    chain = (
      {
        'context': self.vect_store.as_retriever() | _format_docs,
        'question': runnables.RunnablePassthrough(),
      }
      | self.prompt
      | self.llm
      | self.output_parser
    )
    result = chain.invoke(text)
    results = []
    for r in result:
      results.append([*r.values()])
    return gaarf.report.GaarfReport(
      results=results, column_names=[*result[0].keys()]
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides HTTP endpoint for trend analysis."""

import fastapi
from langchain_google_vertexai import ChatVertexAI

from trend_analyzer import analyzer, vectorstore

app = fastapi.FastAPI()
llm = ChatVertexAI(model='gemini-1.5-flash-001')
trend_analyzer = analyzer.TrendsAnalyzer(
  vect_store=vectorstore.load_vectorstore(), llm=llm
)


@app.post('/')
def categorize(
  data: dict[str, str] = fastapi.Body(embed=True),
) -> list[dict[str, str | float]]:
  """Performs text categorization.

  Args:
    data: Contains texts for categorization.

  Returns:
    Category for each text.
  """
  return trend_analyzer.analyze(data.get('question')).to_list(row_type='dict')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining vector store.

Vectorstore contains necessary data to improve LLM results.
"""

import itertools
import pathlib

import chromadb
import langchain_chroma
import langchain_text_splitters
from langchain_community import document_loaders, embeddings
from langchain_core import vectorstores


def get_vector_store(
  vectorstore_type: vectorstores.VectorStore = langchain_chroma.Chroma,
  vector_store_directory: str = './../assistant_db',
  embedding_function=embeddings.FakeEmbeddings(size=1352),
):
  """Loads or creates vectorstore."""
  if not vector_store_directory:
    return load_vectorstore(
      vectorstore_type, vector_store_directory, embedding_function
    )
  return vectorstore_type(
    persist_directory=vector_store_directory,
    embedding_function=embedding_function,
    client_settings=chromadb.config.Settings(anonymized_telemetry=False),
  )


def load_vectorstore(
  vectorstore_type: vectorstores.VectorStore = langchain_chroma.Chroma,
  samples_directory: pathlib.Path = pathlib.Path(__file__).resolve().parent
  / 'files',
  embedding_function=embeddings.FakeEmbeddings(size=1352),
) -> langchain_chroma.Chroma:
  """Creates vectorstore from documents and embeddings."""
  splits = []
  for file in samples_directory.iterdir():
    if file.is_file():
      loader = document_loaders.JSONLoader(
        file_path=file, jq_schema='.results', text_content=False
      )
      docs = loader.load()

      text_splitter = (
        langchain_text_splitters.character.RecursiveCharacterTextSplitter(
          chunk_size=1000, chunk_overlap=200
        )
      )
    splits.append(text_splitter.split_documents(docs))
  return vectorstore_type.from_documents(
    documents=itertools.chain.from_iterable(splits),
    embedding=embedding_function,
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper for importing package."""

import pathlib

import setuptools

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setuptools.setup(
  name='trend-analyzer',
  version='0.1.1',
  python_requires='>3.11',
  description='Library for performing trends analysis via LLMs.',
  long_description=README,
  long_description_content_type='text/markdown',
  author='Google Inc. (gTech gPS CSE team)',
  author_email='no-reply@google.com',
  license='Apache 2.0',
  classifiers=[
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.11',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: Apache Software License',
  ],
  packages=setuptools.find_packages(),
  package_data={'': ['*.txt', 'files/*.json']},
  install_requires=[
    'fastapi==0.111.0',
    'smart_open',
    'google-ads-api-report-fetcher[bq]==1.14.3',
    'langchain==0.2.7',
    'langchain_chroma==0.1.2',
    'langchain-core==0.2.21',
    'langchain-community==0.2.7',
    'langchain-google-genai==1.0.7',
    'langchain-google-vertexai',
    'chromadb==0.5.5',
    'db-dtypes==1.2.0',
    'jq',
  ],
  setup_requires=['pytest-runner'],
  tests_requires=['pytest', 'pytest-mock', 'pandas'],
  entry_points={
    'console_scripts': [
      'trend-analyzer=entrypoints.cli:main',
    ]
  },
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exposes trend analyzer as a tool for Langchain agents."""

import langchain_core

from trend_analyzer import analyzer, llms, vectorstore


class TrendsAnalyzerInput(langchain_core.pydantic_v1.BaseModel):
  """Input for text categorization."""

  question: str = langchain_core.pydantic_v1.Field(
    description='trend analysis question'
  )


class TrendsAnalyzerResults(langchain_core.tools.BaseTool):
  """Tools that performs trend analysis.

  Attributes:
    llm_type: Type of LLM to use for categorization.
    llm_parameters: Parameter for LLM initialization.
  """

  llm_type: str = 'gemini'
  llm_parameters: dict[str, str] = {'model': 'gemini-1.5-flash'}
  name: str = 'trend_analyzer_results_json'
  description: str = 'infers trends based on a vectorstore documents'
  args_schema: type[langchain_core.pydantic_v1.BaseModel] = TrendsAnalyzerInput

  def _run(
    self,
    question: str,
  ) -> list[dict[str, str]]:
    """Performs trend analysis based on LLM and vectorstore.

    Args:
      question: Question to LLM.

    Returns:
      Mappings between text and its category.
    """
    llm = llms.create_trend_analyzer_llm(self.llm_type, self.llm_parameters)
    trend_analyzer = analyzer.TrendsAnalyzer(
      llm=llm, vect_store=vectorstore.load_vectorstore()
    )
    return trend_analyzer.analyze(question).to_list(row_type='dict')
